import 'dart:async';
import 'dart:io';

import 'package:flex/flex.dart';

/// The type of code represented by [AnsiCode].
class AnsiCodeType {
  const AnsiCodeType._(this.name);

  /// A foreground color.
  static const AnsiCodeType foreground = AnsiCodeType._('foreground');

  /// A style.
  static const AnsiCodeType style = AnsiCodeType._('style');

  /// A background color.
  static const AnsiCodeType background = AnsiCodeType._('background');

  /// No formatting.
  static const AnsiCodeType none = AnsiCodeType._('none');

  /// The name of this type of ansi code
  final String name;

  @override
  String toString() => 'AnsiCodeType.$name';
}

/// Standard ANSI escape code for customizing terminal text output.
///
/// [Source](https://en.wikipedia.org/wiki/ANSI_escape_code#Colors)
class AnsiCode {
  /// Creates a new instance of [AnsiCode]
  const AnsiCode(this.name, this.type, this.code, {this.resetCode = 39});

  /// A description of this code.
  final String name;

  /// The type of code that is represented.
  final AnsiCodeType type;

  /// The numeric value associated with this code.
  @nullable
  final String code;

  /// The [AnsiCode] that resets this value.
  ///
  /// Must be specified in constructor if type is [AnsiCodeType.style].
  final int resetCode;

  /// Represents the resetValue escaped for use in terminal output.
  String get reset => code == null ? '' : '$escapeTemplate[${resetCode}m';

  /// Represents the resetValue as an unescaped literal suitable for scripts.
  String get resetForScript => code == null ? '' : '$escapeForScriptTemplate[${resetCode}m';

  /// Provides the appropriate reset value for this [AnsiCode] for the situation, specified by `forScript`.
  String resetValue({bool forScript = false}) => forScript ? resetForScript : reset;

  /// Represents the value escaped for use in terminal output.
  String get escape => code == null ? '' : '$escapeTemplate[${code}m';

  /// Represents the value as an unescaped literal suitable for scripts.
  String get escapeForScript => code == null ? '' : '$escapeForScriptTemplate[${code}m';

  /// Returns the escape value for a script if forScript is true, otherwise returns the escape value for terminal output.
  String escapeValue({bool forScript = false}) {
    return forScript ? escapeForScript : escape;
  }

  /// Wraps [value] with [escape] as a prefix and [resetValue] as a suffix.
  ///
  /// If [forScript] is `true`, the return value contains the "\" neccessary for escaping in a script. The value of [ansiOutputEnabled] is also ignored.
  ///
  /// Returns `value` unchanged if
  ///   * [value] is `null` or empty
  ///   * both [ansiOutputEnabled] and [forScript] are `false`.
  ///   * [type] is [AnsiCodeType.reset]
  String wrap(String value, {bool forScript = false}) => value.isNotEmpty && code != null ? '${escapeValue(forScript: forScript)}$value${resetValue(forScript: forScript)}' : value;

  @override
  String toString() => '$name ${type.name} ($code)';

  /// Whether formatted ANSI output is enabled for [wrapWith] and [AnsiCode.wrap].
  ///
  /// By default, returns `true` if both `stdout.supportsAnsiEscapes` and
  /// `stderr.supportsAnsiEscapes` from `dart:io` are `true`.
  ///
  /// The default can be overridden by setting the [Zone] variable [AnsiCode] to
  /// either `true` or `false`.
  ///
  /// [overrideAnsiOutput] is provided to make this easy.
  static bool get ansiOutputEnabled => Zone.current['AnsiFanci'] ?? (stdout.supportsAnsiEscapes && stderr.supportsAnsiEscapes);

  /// Allows overriding [ansiOutputEnabled] to [enableAnsiOutput] for the code run
  /// within [body].
  static T overrideAnsiOutput<T>(Callback<T> body, {bool enableAnsiOutput = true}) => runZoned<T>(body, zoneValues: {'AnsiFanci': enableAnsiOutput});

  /// Template for ansi escape codes for use in terminals.
  static const escapeTemplate = '\x1B';

  /// Template for ansi escape codes for use in scripts.
  static const escapeForScriptTemplate = r'\033';

  /// [AnsiCode] for light black text
  static const lightBlack = AnsiCode('light black', AnsiCodeType.foreground, '90');

  /// [AnsiCode] for light red text
  static const lightRed = AnsiCode('light red', AnsiCodeType.foreground, '91');

  /// [AnsiCode] for light green text
  static const lightGreen = AnsiCode('light green', AnsiCodeType.foreground, '92');

  /// [AnsiCode] for light yellow text
  static const lightYellow = AnsiCode('light yellow', AnsiCodeType.foreground, '93');

  /// [AnsiCode] for light blue text
  static const lightBlue = AnsiCode('light blue', AnsiCodeType.foreground, '94');

  /// [AnsiCode] for light magenta text
  static const lightMagenta = AnsiCode('light magenta', AnsiCodeType.foreground, '95');

  /// [AnsiCode] for light cyan text
  static const lightCyan = AnsiCode('light cyan', AnsiCodeType.foreground, '96');

  /// [AnsiCode] for light white text
  static const lightWhite = AnsiCode('light white', AnsiCodeType.foreground, '97');

  /// [AnsiCode] for no formatting
  static const none = AnsiCode('none', AnsiCodeType.none, null);

  /// [AnsiCode] for custom orange text
  static const orange = AnsiCode('light orange', AnsiCodeType.foreground, '38;2;191;131;97');

  /// [AnsiCode] for custom diamond text
  static const diamond = AnsiCode('light diamond', AnsiCodeType.foreground, '38;2;179;227;226');
}
