import 'dart:async';
import 'dart:collection' show Queue;
import 'dart:io' as io;

import 'package:logging/logging.dart' as dlog;
import 'package:meta/meta.dart';
import 'package:flex/expose.dart';
import 'package:flex/flex.dart';
import 'package:plexus/plexus.dart';
import 'package:test_api/src/frontend/expect.dart' show TestFailure;

/// There is a single [RootModule] instance, which should **only** be managed (setting level, etc.) by the final product (i.e. Flutter app (not library) or CLI (seperated from api)) and should be accessed statically outside this library.
class RootModule extends Module with ModuleCommon {
  /// Returns the root module, generating it with `authEnabled`, `onInit`, `onDispose` and `level` if not found.
  factory RootModule({bool authEnabled = false, FutureOrCallback onInit = asyncVoidHandler, FutureOrCallback onDispose = asyncVoidHandler, Level level = defaultLevel}) => Module(const [], builder: (_, __) => RootModule._(PatientStreamQueue<Event>(), authEnabled, onInit, onDispose, level), verifyRoot: false);

  RootModule._(this._internalQueue, this.authEnabled, this.onInit, this.onDispose, this.internalLevel) : super.internal() {
    dlog.Logger.root.level = defaultLevel.dlogLevel;
  }

  @override
  RootModule get parent => this;

  @override
  Level internalLevel = defaultLevel;

  /// Whether authentication is enabled for the app
  final bool authEnabled;

  /// Called when [init] is called
  final FutureOrCallback<void> onInit;

  /// Called when [dispose] is called
  final FutureOrCallback<void> onDispose;

  /// The state of the app
  State state = State.uninitialized;

  @protected
  final PatientStreamQueue<Event> _internalQueue;

  @override
  Level get level => internalLevel ?? (level = defaultLevel);

  @override
  set level(Level value) {
    if (value != null) {
      internalLevel = value;
      dlog.Logger.root.level = value.dlogLevel;
    }
  }

  @override
  final UnmodifiableListView<String> package = const UnmodifiableListView([]);

  // void alert(dynamic error, [StackTrace stackTrace]) => tap.addError(error, stackTrace);

  @override
  Stream<Event> get stream => _internalQueue.stream;

  /// The subscriptions registry.
  final Map<String, StreamSubscription<Event>> subscriptions = {};

  @override
  void add(Event event) => _internalQueue.add(event);

  /// Adds an error to the event queue
  void addError(Object error, [StackTrace stackTrace]) => _internalQueue.addError(error, stackTrace);

  /// Safely subscribes to events
  ///
  /// If reference [id] is not already registered, registers the result of [generator].
  /// If specified, [E] can be used to tighten the return type, although casting of the stream itself must be handled by [generator].
  /// Keep in mind that if [E] is specified and [id] is already registered with a type other than [E], a TypeError will be thrown.
  @optionalTypeArgs
  StreamSubscription<E> safeSubscribe<E extends Event>(String id, StreamSubscription<E> Function(String id, PatientStreamQueue<Event> stream) generator) {
    if (subscriptions.containsKey(id)) {
      log(Level.WARNING, 'Subscription with id `$id` already exists. Not added.');
      try {
        return subscriptions[id];
      } on TypeError catch (error, stackTrace) {
        log(Level.ERROR, 'Subscription with id `$id` is a ${subscriptions[id].runtimeType}, not required type $E.', error: TypeError(), stackTrace: stackTrace);
        rethrow;
      }
    }

    return subscriptions[id] = generator(id, _internalQueue);
  }

  @override
  StreamSubscription<Event> subscribe(String id, Value1Callback<Event, void> callback) => safeSubscribe(id, (_, queue) => queue.listen(callback));

  @override
  FutureOr<void> unsubscribe(String id) {
    if (subscriptions.containsKey(id)) {
      final f = subscriptions[id].cancel();
      subscriptions.remove(id);
      return f;
    } else {
      log(Level.WARNING, 'Subscription with id `$id` does not exist. Not removed.');
    }
  }

  @override
  Future<void> unsubscribeAll() {
    final _queue = Queue<MapEntry<String, StreamSubscription<Event>>>.of(subscriptions.entries.where((e) => !e.key.startsWith('__') && !e.key.endsWith('__')));

    final waitFor = <Future<void>>[];

    while (_queue.isNotEmpty) {
      final current = _queue.removeFirst();
      waitFor.add(subscriptions[current].cancel());
      subscriptions.remove(current);
    }

    return Future.wait<void>(waitFor);
  }

  /// Returns a [StreamSubscription] of messages added to this [Logger] that pass [test].
  ///
  /// You can listen for messages using the standard stream APIs, for instance:
  ///
  /// ```dart
  /// logger.registerHandler((event) { ... });
  /// ```
  @override
  StreamSubscription<Event> where(String id, Value1Callback<Event, void> callback, ValueTest<Event> test) => safeSubscribe(id, (_, queue) => queue.where(test).listen((e) => callback(e)));

  @override
  StreamSubscription<E> whereType<E extends Event>(String id, Value1Callback<E, void> callback) => safeSubscribe(id, (_, queue) => queue.where((e) => e is E).cast<E>().listen((e) => callback(e)));

  void _kernelEventHandler() {
    /// Terminal interrupt (usually Ctrl-C) -- Can be caught or ignored by any process | Useful for performing cleanup
    io.ProcessSignal.sigint.watch().listen((_) => dispose(wait: true, exitCode: 128 + 2));

    /// Terminal quit -- Sent by a terminal to indicate that the process should quit and perform a core dump
    // io.ProcessSignal.sigquit.watch().listen((_) => dispose(wait: true, exitCode: 128 + 3));

    /// Kill -- can't be caught or ignored by any process (except the init process on *Nix-based systems)
    // io.ProcessSignal.sigkill.watch().listen((_) => dispose(wait: false, exitCode: 128 + 9));

    /// Requested termination -- Can be caught or ignored by any process | Useful for performing cleanup
    io.ProcessSignal.sigterm.watch().listen((_) => dispose(wait: false, exitCode: 128 + 15));
  }

  // Future<void> _authEventHandler(AuthEvent event) async {
  //   @nullable
  //   Object _error;
  //   @nullable
  //   StackTrace _stackTrace;

  //   if (event.signal == AuthSignal.SignedOut || event.signal == AuthSignal.SignInFailed) {
  //     _stackTrace = event.stackTrace ?? StackTrace.current;
  //     _error = event.error; //?? 'Sorry, no error provided.';
  //   }

  //   switch (event.signal) {
  //     case AuthSignal.SignedOut:
  //       log(Level.ERROR, 'Sign In Failed.', longMessage: 'Sign in failed for some reason. This is the default SignInFailed handler. Maybe? try again.', error: _error, stackTrace: _stackTrace);
  //       break;
  //     case AuthSignal.SignInFailed:
  //       log(Level.ERROR, 'Sign In Failed.', longMessage: 'Sign in failed for some reason. This is the default SignInFailed handler. Maybe? try again.', error: _error, stackTrace: _stackTrace);
  //       break;
  //     case AuthSignal.SignedIn:
  //       log(Level.SUCCESS, 'We\'re Signed In!');
  //       break;
  //     case AuthSignal.SignOutFailed:
  //       break;
  //   }
  // }

  /// Start the app.
  ///
  /// **MUST be called by the end user/app. NOT by libraries/packages**
  ///
  /// Should only be used once in the lifecycle of the app. Using more than once is not tested and is likely to fail.
  Future<void> start(FutureOrCallback<void> main, {bool autoDispose = false}) async {
    if (state == State.uninitialized) {
      await runZoned<Future<void>>(
        () async {
          state = State.initializing;

          try {
            log(Level.FINEST, 'Initializing kernel event handlers...', init: true);
            _kernelEventHandler();
            log(Level.FINER, 'Initialized kernel event handlers.', init: true);
          } catch (error, stackTrace) {
            log(Level.ERROR, 'Error occurred during registration of internal handlers.', error: error, stackTrace: stackTrace, zone: Zone.current);
            rethrow;
          }

          logAuth(Level.FINEST, 'Initial State', AuthSignal.SignedOut);

          // try {
          log(Level.FINEST, 'Initializing Logger...', init: true);
          dlog.Logger.root.onRecord.listen((record) {
            final shortMessage = record.message.length > 80 ? '${record.message.substring(0, 80 - 3)}...' : record.message;
            final longMessage = record.message.length > 80 ? record.message : null;
            add(StdEvent(
              Module(record.loggerName.toLowerCase().replaceAll(r'[^a-z0-9_\.]', '_').split('.')),
              Level.dlogToPlexus[record.level],
              shortMessage,
              longMessage: longMessage,
              error: record.error,
              stackTrace: record.stackTrace,
              blob: record.object,
              timestamp: record.time,
              zone: record.zone,
            ));
          });
          log(Level.FINER, 'Initialized logger.', init: true);
          // } catch (error, stackTrace) {
          //   // log(Level.ERROR, 'Error occurred during initialization of logger.', error: error, stackTrace: stackTrace, zone: Zone.current);
          //   rethrow;
          // }

          try {
            log(Level.FINEST, 'Initializing Plexus Root Module...', init: true);

            await onInit();

            log(Level.FINER, 'Initialized Plexus Root Module.', init: true);
          } catch (error, stackTrace) {
            log(Level.ERROR, 'Error occurred during initialization of Plexus root module.', error: error, stackTrace: stackTrace, zone: Zone.current);
            rethrow;
          }

          state = State.initialized;

          try {
            await main();
          } catch (error, stackTrace) {
            if (error is TestFailure) rethrow;

            log(Level.ERROR, 'Error occurred during execution of main().', error: error, stackTrace: stackTrace, zone: Zone.current);
          }

          if (autoDispose) {
            await dispose();
          }
        },
        zoneSpecification: ZoneSpecification(
          print: (_, __, zone, message) {
            if (message.startsWith(encodedEventHeader)) {
              final event = Event.parse(message, this, zone);
              _internalQueue.isClosed ? io.stdout.writeln('[${event.level}]: ${event.message}') : add(event);
            } else {
              _internalQueue.isClosed ? io.stdout.writeln(message) : log(Level.PRINT, message, zone: zone);
            }
          },
          handleUncaughtError: (_, __, zone, error, stackTrace) {
            // try {
            log(Level.FATAL, 'Unhandled error occurred within the Plexus root module zone. This is most likely due to an error during logger initialization.', error: error, stackTrace: stackTrace, zone: zone);
            // } catch (lError, logStackTrace) {
            //   rethrow;
            // }

            // dispose();
          },
        ),
        onError: (error, stackTrace) {
          log(Level.FATAL, 'Unhandled error occurred within the Plexus root module zone. This is most likely due to an error during logger initialization.', error: error, stackTrace: stackTrace);
        }
      );
    }
  }

  /// Disposes of the root module.
  ///
  /// **MUST be called by the end user/app. NOT by libraries/packages**
  ///
  /// Should only be used once in the lifecycle of the app. Using more than once is not tested and is likely to fail.
  Future<void> dispose({bool wait = true, int exitCode = 0}) async {
    print('Termination in progress...');

    await onDispose();

    if (wait) {
      await _internalQueue.close(wait: wait);
    } else {
      _internalQueue.dispose();
    }

    print('Goodbye.');

    io.exit(exitCode);
  }
}
